/* Solutie la problema Ocolire, concursul Alexandru Rogojan editia 2010.

   Citirea si afisarea se fac de la stdin respectiv stdout. 
   Redirectati corespunzator intrarea si iesirea pentru testare. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#define MAX_LEN 10000

int **alloc_matrix(int n)
{
    int **a, i, j;
    a = (int **) malloc(n * sizeof(int *));
    if (!a)
	return NULL;
    for (i = 0; i < n; i++) {
	a[i] = (int *) calloc(n, sizeof(int));
	if (!a[i]) {
	    for (j = 0; j < i; j++)
		free(a[j]);
	    free(a);
	    return NULL;
	}
    }
    return a;
}

void separate(int *father, int *sep, int n, int node, int faulty,
	      int crossed)
{
    int i;
    int newsep = crossed || (node == faulty);
    if (node >= 0)
	sep[node] = newsep;
    for (i = 0; i < n; i++) {
	if (father[i] == node)
	    separate(father, sep, n, i, faulty, newsep);
    }
}

int main(void)
{
    int n, np, i, j, k, min, imin, max, delta;
    int **a, *path, *sep, *dist, *father, *viz;
    char line[MAX_LEN], *tok;

    // Citire date si alocare memorie pentru calcule.
    fgets(line, MAX_LEN, stdin);
    sscanf(line, "%d", &n);

    a = alloc_matrix(n);
    if (!a) {
	printf("Error at matrix alloc.\n");
	exit(EXIT_FAILURE);
    }

    for (i = 0; i < n; i++) {
	fgets(line, MAX_LEN, stdin);
	tok = strtok(line, " \t\n");
	for (j = 0; j < n; j++) {
	    sscanf(tok, "%d", &a[i][j]);
	    tok = strtok(NULL, " \t\n");
	}
    }

    path = (int *) calloc(n, sizeof(int));
    if (!path) {
	printf("Error at path calloc.\n");
	exit(EXIT_FAILURE);
    }
    fgets(line, MAX_LEN, stdin);
    i = 0;
    tok = strtok(line, " \t\n");
    while (tok) {
	sscanf(tok, "%d", &path[i]);
	path[i]--;
	i++;
	tok = strtok(NULL, " \t\n");
    }
    np = i;

    // Dijkstra de la nodul destinatie la toate celelalte noduri.
    dist = (int *) malloc(n * sizeof(int));
    father = (int *) malloc(n * sizeof(int));
    viz = (int *) calloc(n, sizeof(int));
    if (!dist || !father || !viz) {
	printf("Error at shortest path calloc.\n");
	exit(EXIT_FAILURE);
    }
    for (i = 0; i < n; i++) {
	dist[i] = SHRT_MAX;
	father[i] = -1;
    }
    dist[path[np - 1]] = 0;
    do {
	min = SHRT_MAX;
	imin = -1;
	for (i = 0; i < n; i++) {
	    if (!viz[i] && dist[i] < min) {
		min = dist[i];
		imin = i;
	    }
	}
	if (imin >= 0) {
	    viz[imin] = 1;
	    for (i = 0; i < n; i++) {
		if (!viz[i] && a[imin][i]
		    && dist[imin] + a[imin][i] < dist[i]) {
		    dist[i] = dist[imin] + a[imin][i];
		    father[i] = imin;
		}
	    }
	}
    } while (imin >= 0);

    // Eliminam pe rand cate o muchie din traseu si gasim costul minim
    // de ocolire pentru ea. Pastram maximul din toate costurile.
    // Complexitate O(n*n*l) unde l e lungimea traseului ales initial.
    sep = (int *) malloc(n * sizeof(int));
    if (!sep) {
	printf("Error at sep calloc.\n");
	exit(EXIT_FAILURE);
    }
    max = 0;
    for (k = np - 2; k >= 0; k--) {
	// Intrerupere la path[k] -- path[k+1].
	// Nodul path[k] si tot subarborele sau din parcurgerea cu 
	// Dijkstra se "separa" de restul grafului.
	separate(father, sep, n, -1, path[k], 0);

	// Pentru ocolire se va trece obligatoriu peste o muchie care 
	// leaga un nod "separat" de un nod "neseparat".
	min = SHRT_MAX;
	for (i = 0; i < n; i++) {
	    for (j = 0; j < n; j++) {
		if (a[i][j] && sep[i] && !sep[j] && (father[i] != j)) {
		    delta =
			dist[i] + dist[j] + a[i][j] - 2 * dist[path[k]];
		    if (min > delta)
			min = delta;
		}
	    }
	}

	if (min > max)
	    max = min;
    }

    // Afisare rezultat.
    printf("%d\n", max);

    // Eliberare resurse.
    for (i = 0; i < n; i++)
	free(a[i]);
    free(a);
    free(path);
    free(sep);
    free(dist);
    free(father);
    free(viz);

    exit(EXIT_SUCCESS);
}
